import argparse
import shutil
import sys
from datetime import datetime
from pathlib import Path

import nbformat
from nbparameterise import extract_parameters, parameter_values, replace_definitions

from cal_tools.tools import CalibrationMetadata
from .calibrate import (
    JobChain, SlurmOptions, run_finalize, get_pycalib_version,
)
from .settings import temp_path

def update_notebooks_params(working_dir: Path, new_values):
    """Modify parameters in copied notebooks"""
    for nb_file in working_dir.glob('*.ipynb'):
        nb = nbformat.read(nb_file, as_version=4)
        params = extract_parameters(nb, lang='python')
        params = parameter_values(params, **new_values)
        new_nb = replace_definitions(nb, params, execute=False, lang='python')
        nbformat.write(new_nb, nb_file)

def new_report_path(old_out_folder, old_report_path, new_out_folder):
    """Update report path if it was in the old output folder"""
    try:
        report_in_output = Path(old_report_path).relative_to(old_out_folder)
    except ValueError:
        return old_report_path
    else:
        return str(Path(new_out_folder, report_in_output))

def main(argv=None):
    ap = argparse.ArgumentParser()
    ap.add_argument("from_dir", type=Path, help="A directory containing steps.json")
    ap.add_argument("--python", help="Path to Python executable to run notebooks")
    ap.add_argument("--out-folder", help="Directory to put output data")
    ap.add_argument("--slurm-partition", help="Submit jobs in this Slurm partition")
    ap.add_argument('--no-cluster-job', action="store_true",
                    help="Run notebooks here, not in cluster jobs")
    args = ap.parse_args(argv)

    if not (args.from_dir / 'steps.json').is_file():
        sys.exit(f"Directory {args.from_dir} does not contain steps.json")

    start_time = datetime.now()
    run_uuid = f"t{start_time:%y%m%d_%H%M%S}"

    working_dir = Path(temp_path, f'slurm_out_repeat_{run_uuid}')
    shutil.copytree(
        args.from_dir, working_dir, ignore=shutil.ignore_patterns('slurm-*.out')
    )
    print(f"New working directory: {working_dir}")

    cal_metadata = CalibrationMetadata(working_dir)
    prev_parameters = cal_metadata['calibration-configurations']

    report_path = cal_metadata['report-path']
    out_folder = prev_parameters['out-folder']
    if args.out_folder:
        report_path = new_report_path(out_folder, report_path, args.out_folder)
        out_folder = prev_parameters['out-folder'] = args.out_folder
        update_notebooks_params(working_dir, {'out_folder': out_folder})
        cal_metadata.save()

    # finalize & some notebooks expect yaml metadata in the output folder
    Path(out_folder).mkdir(parents=True, exist_ok=True)
    shutil.copy(working_dir / 'calibration_metadata.yml', out_folder)

    job_chain = JobChain.from_dir(
        working_dir, python=(args.python or sys.executable)
    )
    if args.no_cluster_job:
        job_chain.run_direct()
        joblist = []
    else:
        joblist = job_chain.submit_jobs(SlurmOptions(
            partition=args.slurm_partition,
        ))

    fmt_args = {'run_path': working_dir,
                'out_path': out_folder,
                'version': get_pycalib_version(),
                'report_to': report_path,
                'in_folder': prev_parameters['in-folder'],
                'request_time': start_time.strftime('%Y-%m-%dT%H:%M:%S'),
                'submission_time': start_time.strftime('%Y-%m-%dT%H:%M:%S'),
                }

    joblist.append(run_finalize(
        fmt_args=fmt_args,
        temp_path=working_dir,
        job_list=joblist,
        sequential=args.no_cluster_job,
        partition=args.slurm_partition
    ))

    if any(j is not None for j in joblist):
        print("Submitted the following SLURM jobs: {}".format(",".join(joblist)))


if __name__ == '__main__':
    sys.exit(main())
